/*
 * This file is part of Maliit Plugins
 *
 * Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies). All rights reserved.
 *
 * Contact: Mohammad Anwari <Mohammad.Anwari@nokia.com>
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or other materials
 * provided with the distribution.
 * Neither the name of Nokia Corporation nor the names of its contributors may be
 * used to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "keyareaconverter.h"

#include "models/styleattributes.h"
#include "models/area.h"
#include "models/keyarea.h"
#include "models/key.h"
#include "logic/keyboardloader.h"

#include <QtCore>
#include <QtGui/QGuiApplication>
#include <QtGui/QScreen>
#include <QtQuick>

namespace MaliitKeyboard {
namespace Logic {
namespace Helper {
int calculateVerticalMargin(LayoutHelper::Orientation orientation, Keyboard& kb)
{
    Q_UNUSED(kb);

    const int freeSpace = uiConst->keypadHeight(orientation)
                            - (uiConst->keyHeight(orientation) * uiConst->numberOfRows())
                            - uiConst->keypadTopMargin(orientation);

    const int vMargin = (freeSpace / (uiConst->numberOfRows()-1)) / 2;

    return vMargin;
}

QVector<int> calculateMargins(LayoutHelper::Orientation orientation,
                              Keyboard& kb)
{
    const qreal keyboardWidth = uiConst->windowGeometryRect(qGuiApp->primaryScreen()->orientation()).size().width();

    QVector<int> margins;

    int spaceTakenByKeys = 0;
    int numberOfKeysInRow = 0;
    for (int index = 0; index < kb.keys.count(); ++index) {
        const KeyDescription &desc(kb.key_descriptions.at(index));
        int width = uiConst->keyWidth(orientation, desc.width);
        spaceTakenByKeys += width;
        numberOfKeysInRow++;

        bool at_row_end((index + 1 == kb.keys.count())
                        || (index + 1 < kb.keys.count()
                            && kb.key_descriptions.at(index + 1).row > desc.row));

        if (at_row_end) {
            int marginThisRow = ( (keyboardWidth - spaceTakenByKeys) / (numberOfKeysInRow-1) ) / 2;
            margins.append(marginThisRow);
            spaceTakenByKeys = 0;
        }
    }

    return margins;
}

void UiConstants::calculateNumberOfRows(Keyboard& kb)
{
    if(m_genericStorage->numberOfRows > 0)
        return;

    m_genericStorage->numberOfRows = 0;
    for (int index = 0; index < kb.keys.count(); ++index) {
        const KeyDescription &desc(kb.key_descriptions.at(index));
        bool at_row_end((index + 1 == kb.keys.count())
                        || (index + 1 < kb.keys.count()
                            && kb.key_descriptions.at(index + 1).row > desc.row));

        if (at_row_end)
            m_genericStorage->numberOfRows++;
    }
}

UiConstants* UiConstants::self = 0;

void UiConstants::initUiConstants()
{
    instance();

    if (!m_initialized) {
        const QString g_maliit_ui_constants_qml(MALIIT_KEYBOARD_DATA_DIR "/maliit-ui-constants.qml");

        QQuickView uiConstantsHelperView;
        uiConstantsHelperView.setSource(QUrl::fromLocalFile(g_maliit_ui_constants_qml));
        QQuickItem* uiConstantsHelper = uiConstantsHelperView.rootObject();

        const QRect rLandscape = qGuiApp->primaryScreen()->mapBetween(
                    qGuiApp->primaryScreen()->primaryOrientation(),
                    Qt::LandscapeOrientation,
                    QGuiApplication::primaryScreen()->geometry());

        const QRect rInvertedLandscape = qGuiApp->primaryScreen()->mapBetween(
                    qGuiApp->primaryScreen()->primaryOrientation(),
                    Qt::InvertedLandscapeOrientation,
                    QGuiApplication::primaryScreen()->geometry());

        const QRect rPortrait = qGuiApp->primaryScreen()->mapBetween(
                    qGuiApp->primaryScreen()->primaryOrientation(),
                    Qt::PortraitOrientation,
                    QGuiApplication::primaryScreen()->geometry());

        const QRect rInvertedPortrait = qGuiApp->primaryScreen()->mapBetween(
                    qGuiApp->primaryScreen()->primaryOrientation(),
                    Qt::InvertedPortraitOrientation,
                    QGuiApplication::primaryScreen()->geometry());


        // generic

        m_genericStorage->keyAreaBorders = QMargins(uiConstantsHelper->property("key_area_borders").toInt(),
                                                     uiConstantsHelper->property("key_area_borders").toInt(),
                                                     uiConstantsHelper->property("key_area_borders").toInt(),
                                                     uiConstantsHelper->property("key_area_borders").toInt()
                                                     );

        m_genericStorage->fontColor = uiConstantsHelper->property("font_color").toByteArray();
        m_genericStorage->fontFamily = uiConstantsHelper->property("font_family").toByteArray();

        // portrait

        qreal portraitHeightRatio = uiConstantsHelper->property("phone_keyboard_height_portrait").toReal();
        if (qGuiApp->primaryScreen()->primaryOrientation() == Qt::LandscapeOrientation)
            portraitHeightRatio = uiConstantsHelper->property("tablet_keyboard_height_portrait").toReal();

        m_portraitStorage->invisibleTouchAreaHeight = uiConstantsHelper->property("portrait_invisible_touch_area").toInt();

        m_portraitStorage->keypadHeight = rPortrait.height() * portraitHeightRatio;
        m_portraitStorage->keyboardScreenWidthRatio = uiConstantsHelper->property("portrait_keyboard_screen_width_ratio").toReal();

        const int portraitKeyboardWidth = rPortrait.width() * m_portraitStorage->keyboardScreenWidthRatio;
        const int xp = (rPortrait.width() - (portraitKeyboardWidth)) / 2;

        // point of origin differs when primary orientation is different
        int yp = 0;
        if(qGuiApp->primaryScreen()->primaryOrientation() == Qt::PortraitOrientation)
            yp = rPortrait.height() - (m_portraitStorage->keypadHeight + m_portraitStorage->invisibleTouchAreaHeight);

        m_portraitStorage->windowGeometryRect = QRect(xp, yp, portraitKeyboardWidth,
                                                      m_portraitStorage->keypadHeight
                                                    + m_portraitStorage->invisibleTouchAreaHeight);

        const int invertedPortraitKeyboardWidth = rInvertedPortrait.width() * m_portraitStorage->keyboardScreenWidthRatio;
        const int xpi = (rInvertedPortrait.width() - (portraitKeyboardWidth)) / 2;

        // point of origin differs when primary orientation is different
        int ypi = 0;
        if(qGuiApp->primaryScreen()->primaryOrientation() == Qt::LandscapeOrientation)
            ypi = rInvertedPortrait.height() - (m_portraitStorage->keypadHeight + m_portraitStorage->invisibleTouchAreaHeight);

        m_portraitStorage->windowGeometryRectInverted = QRect(xpi, ypi, invertedPortraitKeyboardWidth,
                                                              m_portraitStorage->keypadHeight
                                                            + m_portraitStorage->invisibleTouchAreaHeight);

        m_portraitStorage->keypadTopMargin = uiConstantsHelper->property("portrait_top_margin").toInt();
        m_landscapeStorage->keypadTopMargin = 0.02 * rLandscape.height(); //uiConstantsHelper->property("landscape_top_margin").toInt();

        m_portraitStorage->keyboardBottomMargin = uiConstantsHelper->property("portrait_bottom_margin").toInt();
        m_landscapeStorage->keyboardBottomMargin = 0.02 * rLandscape.height();//uiConstantsHelper->property("landscape_bottom_margin").toInt();

        // landscape

        qreal landscapeHeightRatio = uiConstantsHelper->property("phone_keyboard_height_landscape").toReal();
        if (qGuiApp->primaryScreen()->primaryOrientation() == Qt::LandscapeOrientation)
            landscapeHeightRatio = uiConstantsHelper->property("tablet_keyboard_height_landscape").toReal();

        m_landscapeStorage->invisibleTouchAreaHeight = uiConstantsHelper->property("landscape_invisible_touch_area").toInt();

        m_landscapeStorage->keypadHeight = rLandscape.height() * landscapeHeightRatio;
        m_landscapeStorage->keyboardScreenWidthRatio = uiConstantsHelper->property("landscape_keyboard_screen_width_ratio").toReal();

        const int landscapeKeyboardWidth = rLandscape.width() * m_landscapeStorage->keyboardScreenWidthRatio;
        const int xl = (rLandscape.width() - (landscapeKeyboardWidth)) / 2;
        const int yl = rLandscape.height() - (m_landscapeStorage->keypadHeight + m_landscapeStorage->invisibleTouchAreaHeight);

        m_landscapeStorage->windowGeometryRect = QRect(xl, yl, landscapeKeyboardWidth,
                                                       m_landscapeStorage->keypadHeight
                                                     + m_landscapeStorage->invisibleTouchAreaHeight);

        const int invertedLandscapeKeyboardWidth = rInvertedLandscape.width() * m_landscapeStorage->keyboardScreenWidthRatio;
        const int xli = (rInvertedLandscape.width() - (invertedLandscapeKeyboardWidth)) / 2;
        const int yli = 0;

        m_landscapeStorage->windowGeometryRectInverted = QRect(xli, yli, invertedLandscapeKeyboardWidth,
                                                               m_landscapeStorage->keypadHeight
                                                             + m_landscapeStorage->invisibleTouchAreaHeight);

        // reference
        const QSize referenceSize(
                    uiConstantsHelper->property("reference_width").toInt(),
                    uiConstantsHelper->property("reference_height").toInt());

        m_landscapeStorage->widthStretchFactor = (qreal) m_landscapeStorage->windowGeometryRect.width() / referenceSize.width();
        m_portraitStorage->widthStretchFactor = (qreal) m_portraitStorage->windowGeometryRect.width() / referenceSize.width();

        m_genericStorage->keyHeight = uiConstantsHelper->property("key_height").toReal();
        m_genericStorage->keyWidthSmall = uiConstantsHelper->property("key_width_small").toReal();
        m_genericStorage->keyWidthMedium = uiConstantsHelper->property("key_width_medium").toReal();
        m_genericStorage->keyWidthLarge = uiConstantsHelper->property("key_width_large").toReal();
        m_genericStorage->keyWidthXLarge = uiConstantsHelper->property("key_width_xlarge").toReal();
        m_genericStorage->keyWidthXXLarge = uiConstantsHelper->property("key_width_xxlarge").toReal();


        m_landscapeStorage->spaceBetweenRows = uiConstantsHelper->property("landscape_space_between_rows").toReal();
        m_landscapeStorage->spaceBetweenKeys = uiConstantsHelper->property("landscape_space_between_keys").toReal();

        m_portraitStorage->spaceBetweenRows = uiConstantsHelper->property("portrait_space_between_rows").toReal();
        m_portraitStorage->spaceBetweenKeys = uiConstantsHelper->property("portrait_space_between_keys").toReal();

        // S M L XL XXL Stretched

        m_genericStorage->fontSize = uiConstantsHelper->property("font_size").toReal();

        m_genericStorage->keyBackgroundNormal = uiConstantsHelper->property("key_background_normal").toByteArray();
        m_genericStorage->keyBackgroundSpecial = uiConstantsHelper->property("key_background_special").toByteArray();
        m_genericStorage->keyBackgroundDead = uiConstantsHelper->property("key_background_deadkey").toByteArray();

        m_initialized = true;

        qDebug() << "landscape         " << rLandscape << m_landscapeStorage->windowGeometryRect;
        qDebug() << "landscape inverted" << rInvertedLandscape << m_landscapeStorage->windowGeometryRectInverted;

        qDebug() << "portrait          " << rPortrait << m_portraitStorage->windowGeometryRect;
        qDebug() << "portrait inverted " << rInvertedPortrait << m_portraitStorage->windowGeometryRectInverted;
    }
}

QRect& UiConstants::windowGeometryRect(Qt::ScreenOrientation orientation)
{
    LayoutHelper::Orientation maliitOrientation = LayoutHelper::Landscape;

    if (orientation == Qt::PortraitOrientation || orientation == Qt::InvertedPortraitOrientation)
            maliitOrientation = LayoutHelper::Portrait;

    if (orientation == Qt::InvertedLandscapeOrientation || orientation == Qt::InvertedPortraitOrientation)
        return storage(maliitOrientation)->windowGeometryRectInverted;

    return storage(maliitOrientation)->windowGeometryRect;
}

qreal UiConstants::keyWidth(LayoutHelper::Orientation orientation, KeyDescription::Width requestedSize)
{
    qreal width = 1.0;

    switch (requestedSize) {
    case KeyDescription::Small:
        width = m_genericStorage->keyWidthSmall;
        break;
    case KeyDescription::Medium:
        width = m_genericStorage->keyWidthMedium;
        break;
    case KeyDescription::Large:
        width = m_genericStorage->keyWidthLarge;
        break;
    case KeyDescription::XLarge:
        width = m_genericStorage->keyWidthXLarge;
        break;
    case KeyDescription::XXLarge:
        width = m_genericStorage->keyWidthXXLarge;
        break;
    case KeyDescription::Stretched:
        width = m_genericStorage->keyWidthStretched;
        break;
    default: break;
    }

    width *= widthStretchFactor(orientation);

    return width;
}

/**
 *  gaps are defined in dp or gu, while key height stretches
 */

qreal UiConstants::keyHeight(LayoutHelper::Orientation orientation)
{
    const qreal numberOfgaps = m_genericStorage->numberOfRows - 1;

    switch (orientation) {
    case LayoutHelper::Landscape:
    {
        const qreal gapSize = m_landscapeStorage->spaceBetweenRows;
        const qreal bottomMargin = m_landscapeStorage->keyboardBottomMargin;
        const qreal topMargin = m_landscapeStorage->keypadTopMargin;
        const qreal keyHeight = (m_landscapeStorage->keypadHeight
                           - (numberOfgaps * gapSize)
                           - bottomMargin
                           - topMargin)
                           / (qreal) m_genericStorage->numberOfRows;
        return keyHeight;
    }
        break;
    case LayoutHelper::Portrait:
    {
        const qreal gapSize = m_portraitStorage->spaceBetweenRows;
        const qreal bottomMargin = m_portraitStorage->keyboardBottomMargin;
        const qreal topMargin = m_portraitStorage->keypadTopMargin;
        const qreal keyHeight = (m_portraitStorage->keypadHeight
                           - (numberOfgaps * gapSize)
                           - bottomMargin
                           - topMargin)
                           / (qreal) m_genericStorage->numberOfRows;
        return keyHeight;
    }
        break;
    default: return 1.0;
    }
}

QByteArray UiConstants::keyBackground(Key::Style style, KeyDescription::State state) const
{
    Q_UNUSED(state);

    switch (style) {
    case Key::StyleNormalKey:
        return m_genericStorage->keyBackgroundNormal;
    case Key::StyleSpecialKey:
        return m_genericStorage->keyBackgroundSpecial;
    case Key::StyleDeadKey:
        return m_genericStorage->keyBackgroundDead;
    default: break;
    }

    return "";
}

Logic::LayoutHelper::Orientation UiConstants::screenToMaliitOrientation(Qt::ScreenOrientation screenOrientation) const
{
    switch (screenOrientation) {
    case Qt::LandscapeOrientation:
    case Qt::InvertedLandscapeOrientation:
        return Logic::LayoutHelper::Landscape;
        break;
    case Qt::PortraitOrientation:
    case Qt::InvertedPortraitOrientation:
    case Qt::PrimaryOrientation:
    default:
        return Logic::LayoutHelper::Portrait;
    }

    return Logic::LayoutHelper::Portrait;
}


} // namespace Helper

//! \class KeyAreaConverter
//! Reads keyboard layouts and converts them into key areas. It can apply
//! styling attributes, making the key areas ready for rendering.
//! Explicit functions are used for the different key areas, instead of a
//! generic lookup. This makes the class easier to use and better documents
//! the valid combinations.

namespace {
//! \brief Creates a key area from a keyboard.
//! \param attributes The styling attributes that get applied to the key area.
//! \param source The keyboard layout used for the key area.
//! \param orientation The layout orientation.
//! \param is_extended_keyarea Whether the resulting key area is used for
//!        extended keys (optional).
KeyArea createFromKeyboard(StyleAttributes *attributes,
                           const Keyboard &source,
                           LayoutHelper::Orientation orientation,
                           bool is_extended_keyarea = false)
{
    Helper::UiConstants::instance()->initUiConstants();

    // An ad-hoc geometry updater that also uses styling information.
    KeyArea ka;
    Keyboard kb(source);

    uiConst->calculateNumberOfRows(kb);

    if (not attributes) {
        qCritical() << __PRETTY_FUNCTION__
                    << "No style given, aborting.";
        return ka;
    }

    attributes->setStyleName(kb.style_name);

    Font font;
    font.setName(uiConst->fontFamily());
    font.setSize(uiConst->fontSize(orientation));
    font.setColor(uiConst->fontColor());

    Font small_font(font);
    small_font.setSize(attributes->smallFontSize(orientation));

    static const QMargins bg_margins( uiConst->keyAreaBorders() );

    const qreal max_width( uiConst->windowGeometryRect(qGuiApp->primaryScreen()->orientation()).width() );
    const qreal key_height(uiConst->keyHeight(orientation));
    const qreal keyboard_visible_height = uiConst->keypadHeight(orientation);

    const qreal padding = attributes->keyAreaPadding(orientation);



    QPoint pos(0, 0);
    QVector<int> row_indices;
    int spacer_count = 0;
    qreal consumed_width = 0;

    QVector<int> margins = Helper::calculateMargins(orientation, kb);
    qreal margin = margins[0];

    int row = 0;
    for (int index = 0; index < kb.keys.count(); ++index) {
        row_indices.append(index);
        Key &key(kb.keys[index]);
        const KeyDescription &desc(kb.key_descriptions.at(index));

        int width = 0;
        pos.setY((key_height + uiConst->spaceBetweenRows(orientation)) * desc.row + uiConst->keypadTopMargin(orientation));
        bool at_row_start((index == 0)
                          || (kb.key_descriptions.at(index - 1).row < desc.row));
        bool at_row_end((index + 1 == kb.keys.count())
                        || (index + 1 < kb.keys.count()
                            && kb.key_descriptions.at(index + 1).row > desc.row));

        if (at_row_end)
            row++;

        if (desc.left_spacer || desc.right_spacer) {
            ++spacer_count;
        }

        width = uiConst->keyWidth(orientation, desc.width);

        const qreal key_margin((at_row_start || at_row_end) ? margin + padding : margin * 2);

        Area area;
        area.setBackground(uiConst->keyBackground(key.style(), KeyDescription::NormalState));
        area.setBackgroundBorders(bg_margins);
        area.setSize(QSize(width + key_margin, key_height));
        key.setArea(area);

        key.setOrigin(pos);
        key.setMargins(QMargins(at_row_start ? padding : margin, margin,
                                at_row_end   ? padding : margin, margin));

        const QString &text(key.label().text());
        key.rLabel().setFont(text.count() > 1 ? small_font : font);

        if (key.icon().isEmpty()) {
            key.setIcon(attributes->icon(desc.icon,
                                         KeyDescription::NormalState));
        } else {
            key.setIcon(attributes->customIcon(key.icon()));
        }

        pos.rx() += key.rect().width();

        if (at_row_end) {
            if (not is_extended_keyarea
                && spacer_count > 0 && pos.x() < max_width + 1) {
                const int spacer_width = qMax<int>(0, max_width - pos.x()) / spacer_count;
                pos.setX(0);
                int right_x = 0;

                Q_FOREACH (int row_index, row_indices) {
                    Key &k(kb.keys[row_index]);
                    const KeyDescription &d(kb.key_descriptions.at(row_index));

                    QRect r(k.rect());
                    QMargins m(k.margins());
                    int extra_width = 0;

                    if (d.left_spacer) {
                        m.setLeft(m.left() + spacer_width);
                        extra_width += spacer_width;
                    }

                    if (d.right_spacer) {
                        m.setRight(m.right() + spacer_width);
                        extra_width += spacer_width;
                    }

                    k.setMargins(m);

                    r.translate(right_x - r.left(), 0);
                    r.setWidth(r.width() + extra_width);
                    k.setOrigin(r.topLeft());
                    k.rArea().setSize(r.size());

                    right_x = r.x() + r.width();
                }
            }

            consumed_width = qMax<qreal>(consumed_width, key.rect().x()
                                                         + key.rect().width()
                                                         + padding);

            row_indices.clear();
            pos.setX(0);
            spacer_count = 0;
        }
    }

    Area area;
    area.setBackground(attributes->keyAreaBackground());
    area.setBackgroundBorders( uiConst->keyAreaBorders() );

    area.setSize(QSize((is_extended_keyarea ? consumed_width : max_width),
                       keyboard_visible_height
                       ));

    ka.setArea(area);
    ka.setOrigin(is_extended_keyarea ? QPoint(0, -attributes->verticalOffset(orientation))
                                     : QPoint(0, attributes->wordRibbonHeight(orientation)));
    ka.setKeys(kb.keys);

    return ka;
}
}


//! \param attributes The styling attributes that should be applied to the
//!                   created key areas.
//! \param loader The keyboard layout loader.
KeyAreaConverter::KeyAreaConverter(StyleAttributes *attributes,
                                   KeyboardLoader *loader)
    : m_attributes(attributes)
    , m_loader(loader)
    , m_orientation(LayoutHelper::Landscape)
{
    if (not attributes || not loader) {
        qFatal("Neither attributes nor loader can be null.");
    }
}


KeyAreaConverter::~KeyAreaConverter()
{}


//! \brief Sets the layout orientation used for creating key areas.
//! \param orientation The layout orientation. Default: landscape.
void KeyAreaConverter::setLayoutOrientation(LayoutHelper::Orientation orientation)
{
    m_orientation = orientation;
}

//! \brief Returns the main key area.
KeyArea KeyAreaConverter::keyArea() const
{
    return createFromKeyboard(m_attributes, m_loader->keyboard(), m_orientation);
}


//! \brief Returns the next key area (right of main key area).
KeyArea KeyAreaConverter::nextKeyArea() const
{
    return createFromKeyboard(m_attributes, m_loader->nextKeyboard(), m_orientation);
}


//! \brief Returns the previous key area (left of main key area).
KeyArea KeyAreaConverter::previousKeyArea() const
{
    return createFromKeyboard(m_attributes, m_loader->previousKeyboard(), m_orientation);
}


//! \brief Returns the main key area with shift bindings activated.
KeyArea KeyAreaConverter::shiftedKeyArea() const
{
    return createFromKeyboard(m_attributes, m_loader->shiftedKeyboard(), m_orientation);
}


//! \brief Returns the symbols key area.
//! \param page The symbols page to return (optional).
KeyArea KeyAreaConverter::symbolsKeyArea(int page) const
{
    return createFromKeyboard(m_attributes, m_loader->symbolsKeyboard(page), m_orientation);
}


//! \brief Returns the main key area with dead key bindings activated.
//! \param dead The key used to look up the dead keys.
KeyArea KeyAreaConverter::deadKeyArea(const Key &dead) const
{
    return createFromKeyboard(m_attributes, m_loader->deadKeyboard(dead), m_orientation);
}


//! \brief Returns the main key area with dead key bindings activated.
//! \param dead The key used to look up the dead keys.
KeyArea KeyAreaConverter::shiftedDeadKeyArea(const Key &dead) const
{
    return createFromKeyboard(m_attributes, m_loader->shiftedDeadKeyboard(dead), m_orientation);
}


//! Returns an extended key area.
//! \param key The key used to look up the extended key binding.
KeyArea KeyAreaConverter::extendedKeyArea(const Key &key) const
{
    return createFromKeyboard(m_attributes, m_loader->extendedKeyboard(key), m_orientation, true);
}


//! Returns the number key area.
KeyArea KeyAreaConverter::numberKeyArea() const
{
    return createFromKeyboard(m_attributes, m_loader->numberKeyboard(), m_orientation);
}


//! Returns the phone number key area.
KeyArea KeyAreaConverter::phoneNumberKeyArea() const
{
    return createFromKeyboard(m_attributes, m_loader->phoneNumberKeyboard(), m_orientation);
}

}} // namespace Logic, MaliitKeyboard
