/*
 * This file is part of Maliit Plugins
 *
 * Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies). All rights reserved.
 *
 * Contact: Mohammad Anwari <Mohammad.Anwari@nokia.com>
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or other materials
 * provided with the distribution.
 * Neither the name of Nokia Corporation nor the names of its contributors may be
 * used to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef MALIIT_KEYBOARD_KEYAREACONVERTER_H
#define MALIIT_KEYBOARD_KEYAREACONVERTER_H

#include "logic/layouthelper.h"
#include "models/keydescription.h"
#include "models/keyboard.h"

#include <QPoint>
#include <QColor>

namespace MaliitKeyboard {

class StyleAttributes;
class KeyboardLoader;
class KeyArea;
class Key;

namespace Logic {

class KeyAreaConverter
{
private:
    StyleAttributes * const m_attributes;
    KeyboardLoader * const m_loader;
    LayoutHelper::Orientation m_orientation;

public:
    explicit KeyAreaConverter(StyleAttributes *attributes,
                              KeyboardLoader *loader);
    virtual ~KeyAreaConverter();

    void setLayoutOrientation(LayoutHelper::Orientation orientation);

    virtual KeyArea keyArea() const;
    virtual KeyArea nextKeyArea() const;
    virtual KeyArea previousKeyArea() const;

    virtual KeyArea shiftedKeyArea() const;
    virtual KeyArea symbolsKeyArea(int page = 0) const;
    virtual KeyArea deadKeyArea(const Key &dead) const;
    virtual KeyArea shiftedDeadKeyArea(const Key &dead) const;
    virtual KeyArea extendedKeyArea(const Key &key) const;
    virtual KeyArea numberKeyArea() const;
    virtual KeyArea phoneNumberKeyArea() const;
};

namespace Helper {
#define uiConst ::MaliitKeyboard::Logic::Helper::UiConstants::instance()

class UiConstants {

    struct UiConstantsStorage {
        UiConstantsStorage()
            : widthStretchFactor(1.0),
              heightStretchFactor(1.0),
              numberOfRows(0)
        {
        }

        int keypadHeight;
        int invisibleTouchAreaHeight;

        QRect windowGeometryRect;
        QRect windowGeometryRectInverted;

        // includes wordribbon
        QRect keyboardRect;

        qreal keyboardScreenWidthRatio;
        qreal keyHeight;
        qreal keyWidthSmall;
        qreal keyWidthMedium;
        qreal keyWidthLarge;
        qreal keyWidthXLarge;
        qreal keyWidthXXLarge;
        qreal keyWidthStretched;
        QMargins keyAreaBorders;
        qreal fontSize;
        QByteArray fontColor;
        QByteArray fontFamily;
        qreal widthStretchFactor;
        qreal heightStretchFactor;
        qreal keypadTopMargin;
        qreal keyboardBottomMargin;

        QByteArray keyBackgroundNormal;
        QByteArray keyBackgroundSpecial;
        QByteArray keyBackgroundDead;

        int numberOfRows;
        qreal spaceBetweenRows;
        qreal spaceBetweenKeys;
    };

public:
    static UiConstants* instance()
    {
        if (!self) {
            self = new UiConstants;
        }
        return self;
    }

    void initUiConstants();
    int keypadHeight(LayoutHelper::Orientation orientation) { return storage(orientation)->keypadHeight; }
    QRect& windowGeometryRect(Qt::ScreenOrientation orientation);
    qreal keyboardScreenWidthRatio(LayoutHelper::Orientation orientation) { return storage(orientation)->keyboardScreenWidthRatio; }

    qreal keyHeight(LayoutHelper::Orientation orientation);
    qreal keyWidth(LayoutHelper::Orientation orientation, KeyDescription::Width width);

    QMargins keyAreaBorders() { return m_genericStorage->keyAreaBorders; }

    qreal fontSize(LayoutHelper::Orientation orientation) { Q_UNUSED(orientation); return m_genericStorage->fontSize; }
    QByteArray fontColor() { return m_genericStorage->fontColor; }
    QByteArray fontFamily() { return m_genericStorage->fontFamily; }

    qreal keypadTopMargin(LayoutHelper::Orientation orientation) { return storage(orientation)->keypadTopMargin; }

    QByteArray keyBackground(Key::Style style, KeyDescription::State state) const;
    Logic::LayoutHelper::Orientation screenToMaliitOrientation(Qt::ScreenOrientation screenOrientation) const;

    qreal spaceBetweenRows(LayoutHelper::Orientation orientation) { return storage(orientation)->spaceBetweenRows; }
    qreal spaceBetweenKeys(LayoutHelper::Orientation orientation) { return storage(orientation)->spaceBetweenKeys; }

    int invisibleTouchAreaHeight(LayoutHelper::Orientation orientation) { return storage(orientation)->invisibleTouchAreaHeight; }

    void calculateNumberOfRows(Keyboard& kb);
    int numberOfRows() const { return m_genericStorage->numberOfRows; }

private:
    UiConstants()
        : m_initialized(false),
          m_portraitStorage(new UiConstantsStorage),
          m_landscapeStorage(new UiConstantsStorage),
          m_genericStorage(new UiConstantsStorage)
    {}
    UiConstants(UiConstants const&){}
    UiConstants& operator=(UiConstants const&){ return *this; }
    static UiConstants* self;

    UiConstantsStorage* storage(LayoutHelper::Orientation orientation)
    {
        initUiConstants();

        if (orientation == LayoutHelper::Landscape)
            return m_landscapeStorage;
        else if (orientation == LayoutHelper::Portrait)
            return m_portraitStorage;

        return m_portraitStorage;
    }

    qreal widthStretchFactor(LayoutHelper::Orientation orientation)
    {
        switch (orientation) {
        case LayoutHelper::Landscape:
            return m_landscapeStorage->widthStretchFactor;
            break;
        case LayoutHelper::Portrait:
            return m_portraitStorage->widthStretchFactor;
            break;
        default: return 1.0;
        }
    }

    qreal heightStretchFactor(LayoutHelper::Orientation orientation)
    {
        switch (orientation) {
        case LayoutHelper::Landscape:
            return m_landscapeStorage->heightStretchFactor;
            break;
        case LayoutHelper::Portrait:
            return m_portraitStorage->heightStretchFactor;
            break;
        default: return 1.0;
        }
    }

    bool m_initialized;

    UiConstantsStorage* m_portraitStorage;
    UiConstantsStorage* m_landscapeStorage;
    UiConstantsStorage* m_genericStorage;
};
}

}} // namespace Logic, MaliitKeyboard

#endif // MALIIT_KEYBOARD_KEYAREACONVERTER_H
